@@ -4,8 +4,14 @@ from fractions import Fraction
 import numpy as np
 import matplotlib.pyplot as plt
 from scipy.stats import chi2
-from scipy.optimize import curve_fit
-from .parameters import flags
+from .parameters import flags, elements
+# warnings.filterwarnings("error")    # To catch warnings with try except.
+# from scipy.optimize import curve_fit
+
+def isotope(name: str, A: int):
+    protons = elements[name]
+    neutrons = A - protons
+    return protons, neutrons
 
 def create_spin_parity_list(
     spins: np.ndarray,
@@ -88,9 +94,9 @@ def gamma_strength_function_average(
     
     TODO: Figure out the pre-factors.
     TODO: Use numpy.logical_or to filter levels and transitions to avoid
+    using many if statements in the loops (less readable, though!).
     TODO: Make res.transitions_BXL.ji, res.transitions_BXL.pii, etc.
     class attributes (properties).
-    using many if statements in the loops.
 
     Parameters
     ----------
@@ -701,8 +707,8 @@ def level_plot(
 
     ax.set_xticks(spin_scope)
     ax.set_xticklabels([f"{Fraction(i)}" + f"$^{parity_symbol}$" for i in spin_scope])
-    ax.set_xlabel("Spin")
-    ax.set_ylabel("E [MeV]")
+    ax.set_xlabel(r"$j^{\pi}$")
+    ax.set_ylabel(r"$E$ [MeV]")
 
     if not ax_input:
         plt.show()
@@ -713,8 +719,9 @@ def level_density(
     include_n_levels: Union[None, int] = None,
     filter_spins: Union[None, int, list] = None,
     filter_parity: Union[None, str, int] = None,
-    E_min: Union[None, float, int] = None,
-    E_max: Union[None, float, int] = None,
+    E_min: Union[float, int] = 0,
+    E_max: Union[float, int] = np.inf,
+    return_counts: bool = False,
     plot: bool = False,
     save_plot: bool = False
     ) -> Tuple[np.ndarray, np.ndarray]:
@@ -745,12 +752,16 @@ def level_density(
         allowed inputs.
 
     E_min : Union[None, float, int]
-        Minimum energy to include in the calculation. If None, the
-        minimum energy in the levels array is used.
+        Minimum energy to include in the calculation.
 
     E_max : Union[None, float, int]
-        Maximum energy to include in the calculation. If None, the
-        maximum energy in the levels array is used.
+        Maximum energy to include in the calculation. If input E_max is
+        larger than the largest E in the data set, E_max is set to that
+        value.
+
+    return_counts : bool
+        Return the counts per bin instead of the density
+        (density = counts/bin_width).
 
     plot : bool
         For toggling plotting on / off.
@@ -790,12 +801,12 @@ def level_density(
         msg = f"'filter_parity' must be of type: None, int, str. Got {type(filter_parity)}."
         raise TypeError(msg)
 
-    if not isinstance(E_min, (type(None), int, float)):
-        msg = f"'E_min' must be of type: None, int, float. Got {type(E_min)}."
+    if not isinstance(E_min, (int, float)):
+        msg = f"'E_min' must be of type: int, float. Got {type(E_min)}."
         raise TypeError(msg)
 
-    if not isinstance(E_max, (type(None), int, float)):
-        msg = f"'E_max' must be of type: None, int, float. Got {type(E_max)}."
+    if not isinstance(E_max, (int, float)):
+        msg = f"'E_max' must be of type: int, float. Got {type(E_max)}."
         raise TypeError(msg)
 
     if isinstance(filter_parity, str):
@@ -821,17 +832,23 @@ def level_density(
         msg = "Parity filter cannot be applied to a list of only energy levels!"
         raise ValueError(msg)
 
-    if levels.ndim == 1:
+    energy_levels = np.copy(levels) # Just in case.
+    if energy_levels.ndim == 1:
         """
-        'levels' only contain energy values.
+        'levels' only contains energy values.
         """
-        energy_levels = levels
-    else:
+        if energy_levels[0] < 0:
+            msg = "Please scale energies relative to the ground state"
+            msg += " energy before calculating the NLD!"
+            raise ValueError(msg)
+
+    elif energy_levels.ndim == 2:
         """
         'levels' is a multidimensional array on the form
-        [[E, 2*spin, parity, idx], ...].
+        [[E, 2*spin, parity, idx], ...]. Subtract ground state energy
+        from all energies.
         """
-        energy_levels = np.copy(levels) # Copy just in case.
+        energy_levels[:, 0] -= energy_levels[0, 0]
         
         if include_n_levels is not None:
             """
@@ -864,54 +881,46 @@ def level_density(
         
         energy_levels = energy_levels[:, 0]
 
-    if levels.ndim == 1:
+    E_max = min(E_max, energy_levels[-1] + 0.1) # E_max cant be larger than the largest energy in the data set. Add small number to include the final level(s) in the counting.
+    if E_max <= E_min:
         """
-        Decide the max value of the energy bins.
+        This behaviour is OK for angular momentum distribution heatmaps
+        where the NLDs for different angular momenta are compared using
+        the same bin range.
+
+        In this situation, the density at E_min is of course zero
+        because there are no levels of this energy for the given angular
+        momentum.
         """
-        if levels[0] != 0:
-            """
-            Calculate energies relative to the ground state if not already
-            done.
-            """
-            energy_levels -= energy_levels[0]
-            bin_max = levels[-1] - levels[0]    # The max energy of the un-filtered data set.
-        else:
-            bin_max = levels[-1]
+        bins = np.array([E_min])
+        density = np.array([0])
+        return bins, density
 
-    else:
-        """
-        Decide the max value of the energy bins.
-        """
-        if levels[0, 0] != 0:
-            """
-            Calculate energies relative to the ground state if not already
-            done.
-            """
-            energy_levels -= energy_levels[0]
-            bin_max = levels[-1, 0] - levels[0, 0]    # The max energy of the un-filtered data set.
-        else:
-            bin_max = levels[-1, 0]
+    bins = np.arange(E_min, E_max + bin_width, bin_width)
+    bins[-1] = E_max    # arange will mess up the final bin if it does not match the bin width.
 
-    if E_min is not None:
-        energy_levels = energy_levels[energy_levels >= E_min]
-
-    if E_max is not None:
-        energy_levels = energy_levels[energy_levels <= E_max]
-
-    bins = np.arange(0, bin_max + bin_width, bin_width)
     n_bins = len(bins)
-    counts = np.zeros(n_bins)
+    counts = np.zeros(n_bins - 1)
 
     for i in range(n_bins - 1):
-        counts[i] = np.sum(bins[i] <= energy_levels[energy_levels < bins[i + 1]])
+        mask_1 = energy_levels >= bins[i]
+        mask_2 = energy_levels < bins[i+1]
+        mask_3 = np.logical_and(mask_1, mask_2)
+        counts[i] = sum(mask_3)
+        # counts[i] = np.sum(bins[i] <= energy_levels[energy_levels < bins[i + 1]])
     
-    density = (counts/bin_width)[:-1]
-    bins = bins[1:]
+    density = (counts/bin_width)
+    # bins = bins[1:]
+    bins = bins[:-1]    # Maybe just a matter of preference...?
 
     if plot:
         fig, ax = plt.subplots()
-        ax.step(bins, density, color="black")
-        ax.set_ylabel(r"Density [MeV$^{-1}$]")
+        if return_counts:
+            ax.step(bins, counts, color="black")
+            ax.set_ylabel(r"Counts")
+        else:
+            ax.step(bins, density, color="black")
+            ax.set_ylabel(r"NLD [MeV$^{-1}$]")
         ax.set_xlabel("E [MeV]")
         ax.legend([f"{bin_width=} MeV"])
         ax.grid()
@@ -921,7 +930,10 @@ def level_density(
             fig.savefig(fname=fname, dpi=300)
         plt.show()
 
-    return bins, density
+    if return_counts:
+        return bins, density, counts
+    else:
+        return bins, density
 
 def porter_thomas(
     transitions: np.ndarray,
@@ -987,7 +999,7 @@ def porter_thomas(
         BXL = transitions[Ei_mask]
     else:
         BXL = transitions[np.abs(transitions[:, 3] - Ei) < Ei_bin_width] # Consider only levels around Ei.
-    
+
     if j_list is not None:
         """
         Create a mask of j values for the transitions array. Allow only
@@ -1028,18 +1040,21 @@ def porter_thomas(
     for ji in initial_j:
         ji_masks.append(BXL[:, 0] == ji)
 
+    n_B_skips = 0
     for pii in pii_masks:
         for idxi in idxi_masks:
             for ji in ji_masks:
                 mask = np.logical_and(ji, np.logical_and(pii, idxi))
                 tmp = BXL[mask][:, 9]   # 9 is B decay.
+                n_B_skips += sum(tmp == 0)  # Count the number of zero values.
+                tmp = tmp[tmp != 0] # Remove all zero values.
                 if not tmp.size:
                     """
                     Some combinations of masks might not match any
                     levels.
                     """
                     continue
-
+                
                 BXL_tmp.extend(tmp/tmp.mean())
 
     BXL = np.asarray(BXL_tmp)
@@ -1047,9 +1062,12 @@ def porter_thomas(
     # BXL = BXL/np.mean(BXL)
     n_BXL_after = len(BXL)
 
-    if n_BXL_before != n_BXL_after:
-        msg = "The number of BXL values has changed during the Porter-Thomas analysis!"
-        msg += f" This should not happen! {n_BXL_before = }, {n_BXL_after = }."
+    if (n_BXL_before - n_B_skips) != n_BXL_after:
+        msg = "The number of BXL values has changed too much during the Porter-Thomas analysis!"
+        msg += f" This should not happen! n_BXL_after should be: {n_BXL_before - n_B_skips}, got: {n_BXL_after}."
+        msg += f"\n{n_BXL_before = }"
+        msg += f"\n{n_BXL_after = }"
+        msg += f"\n{n_B_skips = }"
         raise RuntimeError(msg)
 
     BXL_bins = np.arange(0, BXL[-1] + BXL_bin_width, BXL_bin_width)
@@ -1090,7 +1108,7 @@ def porter_thomas(
     to use the mean of 19 ([1:20] slice, pretty arbitrary chosen) of
     these values to make a more stable normalisation coefficient.
     """
-    BXL_counts *= np.mean(rv.pdf(BXL_bins)[1:20]/BXL_counts[1:20])
+    BXL_counts_normalised = BXL_counts*np.mean(rv.pdf(BXL_bins)[1:20]/BXL_counts[1:20])
     pt_post_process_time = time.perf_counter() - pt_post_process_time
 
     if flags["debug"]:
@@ -1099,16 +1117,18 @@ def porter_thomas(
         print(f"Porter-Thomas: Count time: {pt_count_time:.3f} s")
         print(f"Porter-Thomas: Post process time: {pt_post_process_time:.3f} s")
         print(f"{sum(BXL_counts) = }")
+        print(f"{sum(BXL_counts_normalised) = }")
+        print(f"{n_B_skips = }")
         print(f"{Ei = }")
         print(f"{Ei_bin_width = }")
         print("--------------------------------")
 
     if return_chi2:
-        # return BXL_bins, BXL_counts_normalised, rv.pdf(BXL_bins)
-        return BXL_bins, BXL_counts, rv.pdf(BXL_bins)
+        # return BXL_bins, BXL_counts, rv.pdf(BXL_bins)
+        return BXL_bins, BXL_counts_normalised, rv.pdf(BXL_bins)
     else:
-        # return BXL_bins, BXL_counts_normalised
-        return BXL_bins, BXL_counts
+        # return BXL_bins, BXL_counts
+        return BXL_bins, BXL_counts_normalised
 
 def nuclear_shell_model():
     """
